package com.sb.support;


import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.net.CookieHandler;
import java.net.CookieManager;
import java.net.CookiePolicy;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.streambase.sb.*;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.adapter.*;
import com.streambase.sb.operator.*;
import com.streambase.sb.Timestamp;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
 

/**
 * Output adapters take streams from an application and make them
 * available outside of the application.
 * <p/>
 * The Tibbr REST API responds to commands and does not generate new events by
 * itself, so the Tibbr adapter is modeled as an OutputAdapter taking the command
 * from the input port and emitting the result from processTuple().
 * <p/>
 * Web services may take an arbitrarily long time to respond, so typical use will 
 * be to make the adapter "run in its own thread" via the EventFlow concurrency
 * setting. The adapter implementation does not run in its own thread by design
 * to handle those use cases where explicitly waiting for the response is desirable. 
 * <p/>
 * The REST API does not persist a connection, so this adapter does not implement
 * a control port for explicit connection and disconnection control. If the connection
 * cannot be made on a per-command basis, an error is returned at that time.
 * 
 */
public class TibbrOut extends OutputAdapter implements Parameterizable {

    public static final long serialVersionUID = 1392410649653L;
    private Logger logger;
    // Properties
    private String TibURL;
    private String user;
    private String password;
    private String displayName = "Tibbr Output";
    private Schema inSchema;
    private Schema outSchema;
    private Schema statusSchema;
    private String authToken = null;
    private String rememberToken = null;
    private String clientKey = "SB_Tibbr_Input_Adapter"; //$NON-NLS-1$
    private String myUserId = "";
    private final String charset = "UTF-8"; //$NON-NLS-1$
    private int RESULT_PORT = 0;
    private int STATUS_PORT = 1;
    private List<String> cookies;
    private LogLevel logLevel = LogLevel.INFO;

    /**
     * The constructor is called when the Adapter instance is created, but before the Adapter 
     * is connected to the StreamBase application.
     * 
     * There is one input port for Tibbr API REST commands and two output ports,
     * one for the command result and one for status (including errors)
     * 
     */
    public TibbrOut() {
        super();
        setPortHints(1, 2);
        setDisplayName(displayName);
        setShortDisplayName(this.getClass().getSimpleName());
        setTibURL("");
        setUser("");
        setPassword("");
    }

    /**
     * Typecheck this adapter. The typecheck method is called after the adapter instance is
     * constructed, and once input ports are connected. The adapter should validate its parameters 
     * and throw PropertyTypecheckException (or TypecheckException) if any problems are found. 
     * The message associated with the thrown exceptions will be displayed in StreamBase Studio during authoring,
     * or printed on the console by sbd. Input adapters should set the schema of each output
     * port by calling the setOutputSchema(portNum, schema) method for each output port. 
     * If the adapter needs to change the number of input ports based on parameter values,
     * it should call requireInputPortCount(portCount) at this point. 
     */
    public void typecheck() throws TypecheckException {
        // Confirm input stream has required fields
        ArrayList<Schema.Field> fields = new ArrayList<Schema.Field>();
        fields.add(new Field("Command",CompleteDataType.forString()));
        fields.add(new Field("Text",CompleteDataType.forString()));
        inSchema = new Schema(null,fields);
        if (! inSchema.isSubsetOf(getInputSchema(0))) {
            throw new TypecheckException("Input port 0 must have fields: Command (string), Text (string)");
        }
        requireInputPortCount(1);

        // Schema for output port command result
        fields.clear();
        fields = new ArrayList<Schema.Field>();
        fields.add(new Field("Class",CompleteDataType.forString()));
        fields.add(new Field("Result",CompleteDataType.forString()));
        outSchema = new Schema(null,fields);
        setOutputSchema(RESULT_PORT, outSchema);
        
        // Schema for status/connectivity port output event tuples 
        fields.clear();
        fields = new ArrayList<Schema.Field>();
        fields.add(new Field("Type",CompleteDataType.forString()));
        fields.add(new Field("Object",CompleteDataType.forString()));
        fields.add(new Field("Action",CompleteDataType.forString()));
        fields.add(new Field("Message",CompleteDataType.forString()));
        fields.add(new Field("Tuple",CompleteDataType.forTuple(inSchema)));
        fields.add(new Field("Time",CompleteDataType.forTimestamp()));
        statusSchema = new Schema(null,fields);
        setOutputSchema(STATUS_PORT, statusSchema);

    }

    /**
     * Initialize the adapter. If typecheck succeeds, the init method is called before
     * the StreamBase application is started.
     * 
     * This adapter specifically does not want to use a runnable thread.
     */
    public void init() throws StreamBaseException {
        super.init();
        //
        // Get our own personal logger -- this allows us to tweak the log level without
        // messing with other instances.
        //
        logger = LoggerFactory.getLogger(getName());
        try {
            setLogLevel(logger, getLogLevel());
        } catch (Exception e) {
            getLogger().warn(e.getLocalizedMessage());
        }
    }

    /**
     * Shutdown adapter by cleaning up any resources used (e.g. close an output file
     * if it has been opened). When the application is shutting down, the adapter's shutdown.
     * method will be called first. Once this has returned, all threads should exit and the
     * adapter is considered shutdown.
     */
    public void shutdown() {

    }
    
    /*
     * Utility functions for log messages
     */

    private void debug(String msg) {
    	if (logger.isDebugEnabled()) logger.debug(this.getName()+": "+msg);
    }
    private void debugExStack(Exception e) {
    	if (logger.isDebugEnabled()) {
    		StringWriter sw = new StringWriter();
    		PrintWriter pw = new PrintWriter(sw);
    		e.printStackTrace(pw);
    		logger.debug(this.getName()+": "+sw.toString());
    	}
    }
    
    private void error(String msg) {
    	if (logger.isErrorEnabled()) logger.error(this.getName()+": "+msg);
    }
    
    private void warn(String msg) {
    	if (logger.isWarnEnabled()) logger.warn(this.getName()+": "+msg);
    }

    private void info(String msg) {
    	if (logger.isInfoEnabled()) logger.info(this.getName()+": "+msg);
    }
    
    /*
     * Emit a status tuple
     */
    private void sendStatusEvent(String type, String obj, String action, String msg, Tuple tup) {
    	Tuple stat = statusSchema.createTuple();
		try {
			stat.setString("Type", type);
			stat.setString("Object", obj);
			stat.setString("Action", action);
			stat.setString("Message", msg);
			stat.setTuple("Tuple", tup);
			stat.setTimestamp("Time", Timestamp.now());
			sendOutput(STATUS_PORT, stat);
		} catch (TupleException e) {
			debugExStack(e);
		} catch (StreamBaseException e) {
			debugExStack(e);
		}
    }
    
    @Override
    public void suspend() throws Exception {
    	super.suspend();
    	info("State: Suspended");
    	sendStatusEvent("State","","Suspended","adapter suspended",null);
    }
    
    @Override
    public void resume() throws Exception {
    	super.resume();
    	info("State: Resumed");
    	sendStatusEvent("State","","Resumed","adapter resumed",null);
    }
    
    /** Log In handler
    *     Sets globals:
    *         authToken
    *         cookies
    *     Emits:
    *         User record XML
    */ 
    private String doLogin() {
        String response = null;
        Boolean isError = false;
        
        if (authToken==null) {
            try {
                URL url = new URL( TibURL+"a/users/login.xml"+"?client_key="+clientKey+"&");

                info("Logging in to: "+TibURL);
                 
                HttpURLConnection rc = (HttpURLConnection)url.openConnection();
                rc.setRequestMethod("POST");
                rc.setDoOutput( true );
                rc.setDoInput( true ); 
                rc.setRequestProperty( "Content-Type", "text/xml; charset=utf-8" );
                String reqStr = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><params>" + 
                        "<login>"+user+"</login><password>"+password+"</password>"+
                        "<remember-me type=\"boolean\">true</remember-me>"+
                        "<session_manager>'cookie'</session_manager>"+
                        "</params>";
                debug("Request string: "+reqStr);
                int len = reqStr.length();
                rc.setRequestProperty( "Content-Length", Integer.toString( len ) );
                rc.connect();
                OutputStreamWriter out = new OutputStreamWriter( rc.getOutputStream(), "UTF-8" ); 
                out.write( reqStr, 0, len );
                out.flush();
                debug("Request sent, reading response ");

                InputStream read;
                try {
                    read = rc.getInputStream();
                } catch (IOException e) {
                    read = rc.getErrorStream();
                    isError = true;
                    error(e.toString());
                    debugExStack(e);
                }
                
                // Parse the XML response to get <auth-token>
                DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
                DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
                Document doc = dBuilder.parse(read);
                doc.getDocumentElement().normalize();
                
                TransformerFactory tf = TransformerFactory.newInstance();
                Transformer transformer = tf.newTransformer();
                transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
                StringWriter writer = new StringWriter();
                transformer.transform(new DOMSource(doc), new StreamResult(writer));
                response = writer.getBuffer().toString();
                
                if (isError) {
                	sendStatusEvent("Connection", response, "Log In", "Log In failed", null);
                } else {
                	sendStatusEvent("Connection", response, "Log In", "Log In succeded", null);
	                NodeList nodes = doc.getElementsByTagName("user");
	    
	                cookies = rc.getHeaderFields().get("Set-Cookie");
	                for (int i=0; i < nodes.getLength(); i++) {
	                    debug("cookie: " + cookies.get(i));
	                }
	                rc.disconnect();
	                
	                for (int i = 0; i < nodes.getLength(); i++) {
	                    Node node = nodes.item(i);
	    
	                    if (node.getNodeType() == Node.ELEMENT_NODE) {
	                        Element element = (Element) node;
	                        authToken = getValue("auth-token", element);
	                        rememberToken = getValue("remember-token",element);
	                        myUserId = getValue("id", element);
	                        debug("Token: " + authToken);
	                        debug("RemToken: " + rememberToken);
	                        debug("id   : " + myUserId); 
	                    }
	                }
                }
            } catch (MalformedURLException e) {
                error(e.toString());
                debugExStack(e);
            } catch (IOException e) {
                error(e.toString());
                debugExStack(e);
            } catch (ParserConfigurationException e) {
                error(e.toString());
                debugExStack(e);
            } catch (SAXException e) {
                error(e.toString());
                debugExStack(e);
            } catch (TransformerException e) {
                error(e.toString());
                debugExStack(e);
            }

        } 
        debug("Response: \n"+response);
        return response;
    }
    
    /* General purpose wrapper for all GET messages
     * 
     * @param cmdUrl extension to TibURL for specific command
     * @param params additional URL parameters ready to be appended to the URL, & delimited
     */
    private String doGET(String cmdUrl, String params) {
        String response = null;
        String urlStr;
        try {
            urlStr = TibURL + cmdUrl +
                    "?client_key="+URLEncoder.encode(clientKey, charset) +
                    "&auth_token="+URLEncoder.encode(authToken, charset) + params;
        
            debug("Open url : "+urlStr);
            URL url = new URL( urlStr );
            HttpURLConnection rc = (HttpURLConnection)url.openConnection();
            rc.setRequestMethod("GET");
            // Do as if you're using Chrome 32
            rc.setRequestProperty("User-Agent", "Mozilla/5.0 (Windows NT 6.2; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/32.0.1667.0 Safari/537.36");
            for (String cookie : cookies) {
                rc.addRequestProperty("Cookie", cookie.split(";", 2)[0]);
            }
            debug("Reading response");
            InputStream read;
            try {
                read = rc.getInputStream();
            } catch (IOException e) {
                read = rc.getErrorStream();
                error(e.toString());
                debug(e.getStackTrace().toString());
            }
    
            // Get response
            StringBuilder sb = new StringBuilder();   
            int ch = read.read();
            while ( ch != -1 ) {
              sb.append((char)ch);
              ch = read.read();
            }
            read.close();
            response = sb.toString();

        } catch (UnsupportedEncodingException e1) {
            error(e1.toString());
            debugExStack(e1);
        } catch (MalformedURLException e1) {
            error(e1.toString());
            debugExStack(e1);
        } catch (IOException e1) {
            error(e1.toString());
            debugExStack(e1);
        }
        debug("Response: \n"+response);
        return response;
    }
    
    /* General purpose wrapper for all POST messages
     * 
     * @param cmdUrl extension to TibURL for sepcific command
     * @param reqStr body of request to be POST-ed 
     */
    private String doPOST(String cmdUrl, String reqStr) {
        String response = null;
        try {
            String urlStr = TibURL + cmdUrl + "?client_key="+URLEncoder.encode(clientKey, charset) + "&auth_token="+URLEncoder.encode(authToken, charset);
            URL url = new URL( urlStr );
            debug("Logging in to: "+TibURL);
            HttpURLConnection rc = (HttpURLConnection)url.openConnection();
            rc.setRequestMethod("POST");
            rc.setDoOutput( true );
            rc.setDoInput( true ); 
            rc.setRequestProperty( "Content-Type", "text/xml; charset=utf-8" );

            debug("Request string: "+reqStr);
            int len = reqStr.length();
            rc.setRequestProperty( "Content-Length", Integer.toString( len ) );

            // Connect, Any IOException will be thrown at getOutputStream(); connect() doesn't really.
            rc.connect();
            OutputStreamWriter out = new OutputStreamWriter( rc.getOutputStream(), charset ); 
            out.write( reqStr, 0, len );
            out.flush();
            debug("Request sent, reading response ");
            
            InputStream read;
            try {
                read = rc.getInputStream();
            } catch (IOException e) {
                read = rc.getErrorStream();
                error(e.toString());
                debugExStack(e);
            }
            
            // Parse the XML response to get <auth-token>
            DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
            DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
            Document doc = dBuilder.parse(read);
            doc.getDocumentElement().normalize();
            
            // Convert result to XML text
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
            StringWriter writer = new StringWriter();
            transformer.transform(new DOMSource(doc), new StreamResult(writer));
            response = writer.getBuffer().toString();
            
        } catch (MalformedURLException e) {
            error(e.toString());
            debugExStack(e);
        } catch (IOException e) {
            error(e.toString());
            debugExStack(e);
        } catch (ParserConfigurationException e) {
            error(e.toString());
            debugExStack(e);
        } catch (SAXException e) {
            error(e.toString());
            debugExStack(e);
        } catch (TransformerConfigurationException e) {
            error(e.toString());
            debugExStack(e);
        } catch (TransformerException e) {
            error(e.toString());
            debugExStack(e);
        }
        debug("Response: \n"+response);
        return response;
    }

    /**
     * This method will be called by the StreamBase server for each
     * Tuple given to the adapter to process.
     * 
     * Returns response XML (with no interpretation).
     * 
     * Log In will be attempted before the command if there has been no successful
     * login in the lifetime of the adapter. 
     * 
     * @param inputPort the input port that the tuple is from (ports are zero based)
     * @param tuple the tuple from the given input port
     * @throws StreamBaseException Terminates the application. 
    
     */
    public void processTuple(int inputPort, Tuple tuple)
            throws StreamBaseException {
    	
    	if (isSuspended()) {
    		warn("Command tuple dropped, adapter Suspended; tuple: "+tuple.toString());
    	} else {
            String cmd = tuple.getString("Command");
            String response = null;
            String objClass = null;
            
            CookieHandler.setDefault(new CookieManager(null, CookiePolicy.ACCEPT_ALL));

            // Log In - to get authorization
            if (authToken==null) {
                response = doLogin();
                if (authToken==null) {
                    warn("Log In failed: Cannot execute \'"+cmd+"\'");
                }
            } 
           
            // Execute REST command
            if (authToken!=null) {

                // Publish to my wall
                if (cmd.contentEquals("pub")) { 
                    // Set Content
                    String reqStr = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"+ 
                            "<message>" +
                            "<content>"+tuple.getString("Text")+"</content>" +
                            "<parent-id nil=\"true\"></parent-id>" +
                            "</message>";
                    response = doPOST("a/messages.xml", reqStr);
                }    
                
                // Get my "User" info
                if (cmd.contentEquals("me")) {
                    response = doGET("a/users/find_by_auth_token.xml", "");
                }          

                // Get User record about another user
                // Command: other
                // Text: user-id (a number)
                if (cmd.contentEquals("other")) {
                    response = doGET("a/users/"+tuple.getString("Text")+".xml", "&profile=complete");
                }
                
                // Get users who follow me
                if (cmd.contentEquals("fol")) {
                    response = doGET("a/users/"+myUserId+"/followers.xml", "&page=1&per_page=10");
                }
                
                // Get users I follow
                if (cmd.contentEquals("idol")) {
                    response = doGET("a/users/"+myUserId+"/idols.xml", "&page=1&per_page=10");
                }

                // Get subjects I have created
                if (cmd.contentEquals("subj")) {
                    response = doGET("a/users/"+myUserId+"/subjects.xml", "&page=1&per_page=10");
                }

                // Get subjects I am subscribed to
                if (cmd.contentEquals("subs")) {
                    response = doGET("a/users/"+myUserId+"/subscriptions.xml", "&page=1&per_page=10");
                }

                // Get my unread message count
                if (cmd.contentEquals("unread")) {
                    response = doGET("a/users/"+myUserId+"badges.xml", "");
                }
                
            }
            
            // Emit response tuple
            if (response!=null) {
                Tuple outTuple = outSchema.createTuple();
                // Get Class (asset, channel, link, message, message_filter, role, schedule, subject, subscription-request, user
                objClass = getXMLClass(response);
                outTuple.setString("Class", objClass);
                outTuple.setString("Result", response);
                sendOutput(RESULT_PORT,outTuple);
            } else {
                debug("Null response received for tuple: "+tuple.toString());
            }
    		
    	}

    }
    
    /*
     * Get XML response Object Class
     */
    private String getXMLClass(String xml) {
        String objClass = "";
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder;
		try {
			builder = factory.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(xml));
            Document doc = builder.parse(is);
            objClass = doc.getFirstChild().getNodeName();
		} catch (ParserConfigurationException e) {
			debugExStack(e);
		} catch (SAXException e) {
			debugExStack(e);
		} catch (IOException e) {
			debugExStack(e);
		}
    	return objClass;
    }
    
    /*
     * XML parsing utility method
     * Returns string value of first node from elements matching tag
     * 
     * @param tag name of the elements to inspect
     * @param element parent element
     */
    private static String getValue(String tag, Element element) {
        NodeList nodes = element.getElementsByTagName(tag).item(0).getChildNodes();
        Node node = (Node) nodes.item(0);
        return node.getNodeValue();
        }

    /***************************************************************************************
     * The getter and setter methods provided by the Parameterizable object.               *
     * StreamBase Studio uses them to determine the name and type of each property         *
     * and obviously, to set and get the property values.                                  *
     ***************************************************************************************/

    public void setTibURL(String TibURL) {
    	if (TibURL.length()>0) {
    		if (TibURL.charAt(TibURL.length()-1)!='/') {
    			TibURL+="/";
    		}
    	}
        this.TibURL = TibURL;
    }

    public String getTibURL() {
        return this.TibURL;
    }

    public void setUser(String user) {
        this.user = user;
    }

    public String getUser() {
        return this.user;
    }

    public void setPassword(String password) {
        this.password = password;
    }

    public String getPassword() {
        return this.password;
    }
    
    public LogLevel getLogLevel()
    {
        return this.logLevel;
    }
    
    public void setLogLevel(LogLevel level)
    {
        this.logLevel = level;
    }

}
